@@ -2,23 +2,23 @@
 use std::borrow::{Borrow, Cow};
 use std::collections::{HashMap, HashSet};
 use std::fmt::format;
-use std::net::SocketAddr;
+use std::net::{IpAddr, SocketAddr};
 use std::string::String;
 
-use chrono::{DateTime, Utc};
 use chrono::offset::TimeZone;
+use chrono::{DateTime, Utc};
 use regex::Regex;
 use tokio::io::{
-    AsyncBufRead, AsyncBufReadExt, AsyncRead, AsyncReadExt, AsyncWriteExt, BufStream, copy,
+    copy, AsyncBufRead, AsyncBufReadExt, AsyncRead, AsyncReadExt, AsyncWriteExt, BufStream,
 };
 use tokio::net::{TcpStream, ToSocketAddrs};
 #[cfg(feature = "ftps")]
 use tokio_rustls::{client::TlsStream, rustls::ClientConfig, rustls::ServerName, TlsConnector};
 
-use crate::{cmd, ftp_reply, MODES, REPLY_CODE_LEN, StringExt};
 use crate::cmd::Command;
 use crate::connection::Connection;
 use crate::types::{FileType, FtpError, Result};
+use crate::{cmd, ftp_reply, StringExt, MODES, REPLY_CODE_LEN};
 
 lazy_static::lazy_static! {
     // This regex extracts IP and Port details from PASV command response.
@@ -30,6 +30,7 @@ lazy_static::lazy_static! {
 
     // This regex extracts File size from SIZE command response.
     static ref SIZE_RE: Regex = Regex::new(r"\s+(\d+)\s*$").unwrap();
+    static ref PROT_COMMAND_VALUE: Vec<&'static str> = {vec!["C","E","S","P"]};
 }
 
 pub struct FtpClient {
@@ -44,7 +45,7 @@ pub struct FtpClient {
 }
 
 impl FtpClient {
-    pub fn new(stream: TcpStream) -> Self {
+    fn new(stream: TcpStream) -> Self {
         FtpClient {
             stream: BufStream::new(Connection::Tcp(stream)),
             #[cfg(feature = "ftps")]
@@ -58,7 +59,7 @@ impl FtpClient {
     }
 
     #[cfg(feature = "ftps")]
-    pub fn new_tls_client(stream: TlsStream<TcpStream>) -> Self {
+    fn new_tls_client(stream: TlsStream<TcpStream>) -> Self {
         FtpClient {
             stream: BufStream::new(Connection::Ssl(stream)),
             ssl_cfg: None,
@@ -105,7 +106,7 @@ impl FtpClient {
     /// let conf = ClientConfig::builder().with_safe_defaults().with_root_certificates(root_store).with_no_client_auth();
     /// let domain = ServerName::try_from("www.cert-domain.com").expect("invalid DNS name");
     /// async {
-    ///   let mut ftp_client = FtpClient::connect("172.25.82.139:21").await.unwrap();
+    ///   let mut ftp_client = FtpClient::connect("192.168.32.204:21").await.unwrap();
     ///   let mut ftp_client = ftp_client.into_secure(conf, domain).await.unwrap();
     /// };
     /// ```
@@ -115,8 +116,7 @@ impl FtpClient {
         config: ClientConfig,
         domain: ServerName,
     ) -> Result<FtpClient> {
-        self.send_command(Command::AUTH, Some("TLS")).await?;
-        self.check_response(ftp_reply::AUTH_OK)?;
+        self.exe_auth_tls().await?;
 
         let connector: TlsConnector = std::sync::Arc::new(config.clone()).into();
         let stream = connector
@@ -136,6 +136,91 @@ impl FtpClient {
         Ok(ftps_client)
     }
 
+    #[cfg(feature = "ftps")]
+    async fn exe_auth_tls(mut self) -> Result<()> {
+        self.send_command(Command::AUTH, Some("TLS")).await?;
+        self.check_response_in(&[ftp_reply::AUTH_OK, ftp_reply::SECURITY_MECHANISM_IS_OK])
+    }
+
+    #[cfg(feature = "ftps")]
+    /// Send the AUTH command with the specified mechanism.
+    pub async fn exe_auth(mut self, mechanism: &str) -> Result<u32> {
+        Ok(self.send_command(Command::AUTH, Some(mechanism)).await?)
+    }
+
+    #[cfg(feature = "ftps")]
+    /// PBSZ command. pbsz value: 0 to (2^32)-1 decimal integer.
+    /// * pbsz: Protection Buffer Size
+    pub async fn exec_pbsz(mut self, pbsz: u64) -> Result<()> {
+        if pbsz < 0 || 4294967295 < pbsz {
+            Err(FtpError::InvalidArgument(format!(
+                "Invalid pbsz value. The correct value should be: {} to {}",
+                0, "(2^32)-1"
+            )))
+        }
+        self.send_command(Command::PBSZ, Some(pbsz.to_string().as_str()))
+            .await?;
+        self.check_response_in(&[ftp_reply::COMMAND_OK])?;
+        Ok(())
+    }
+
+    #[cfg(feature = "ftps")]
+    /// Send the ADAT command with the specified authentication data.
+    pub async fn exec_adat(mut self, data: Option<&[u8]>) -> Result<u32> {
+        let mut args = None;
+        if data.is_some() {
+            args = Some(base64::encode(data).as_str());
+        }
+        Ok(self.send_command(Command::ADAT, args).await?)
+    }
+
+    #[cfg(feature = "ftps")]
+    /// Send the CONF command with the specified authentication data.
+    pub async fn exec_conf(mut self, data: Option<&[u8]>) -> Result<u32> {
+        let mut args = None;
+        if data.is_some() {
+            args = Some(base64::encode(data).as_str());
+        }
+        Ok(self.send_command(Command::CONF, args).await?)
+    }
+
+    #[cfg(feature = "ftps")]
+    /// Send the ENC command with the specified authentication data.
+    pub async fn exec_enc(mut self, data: Option<&[u8]>) -> Result<u32> {
+        let mut args = None;
+        if data.is_some() {
+            args = Some(base64::encode(data).as_str());
+        }
+        Ok(self.send_command(Command::ENC, args).await?)
+    }
+
+    #[cfg(feature = "ftps")]
+    /// Send the MIC command with the specified authentication data.
+    pub async fn exec_mic(mut self, data: Option<&[u8]>) -> Result<u32> {
+        let mut args = None;
+        if data.is_some() {
+            args = Some(base64::encode(data).as_str());
+        }
+        Ok(self.send_command(Command::MIC, args).await?)
+    }
+
+    #[cfg(feature = "ftps")]
+    /// Send the MIC command with the specified authentication data.
+    pub async fn exec_prot(mut self, prot: &mut str) -> Result<()> {
+        let mut p = prot.as_mut();
+        if p.is_empty() {
+            p = &mut "C";
+        }
+        if !PROT_COMMAND_VALUE.contains(&&*p) {
+            Err(FtpError::InvalidArgument(format!(
+                "Unsupported prot command value",
+            )))
+        }
+        self.send_command(Command::PROT, Some(p)).await?;
+        self.check_response_in(&[ftp_reply::COMMAND_OK])?;
+        Ok(())
+    }
+
     /// Switch to insecure mode. If the connection is already
     /// insecure does nothing.
     ///
@@ -152,7 +237,7 @@ impl FtpClient {
     /// let conf = ClientConfig::builder().with_safe_defaults().with_root_certificates(root_store).with_no_client_auth();
     /// let domain = ServerName::try_from("www.cert-domain.com").expect("invalid DNS name");
     /// async {
-    ///   let mut ftp_client = FtpClient::connect("172.25.82.139:21").await.unwrap();
+    ///   let mut ftp_client = FtpClient::connect("192.168.32.204:21").await.unwrap();
     ///   let mut ftp_client = ftp_client.into_secure(conf, domain).await.unwrap();
     ///   // Switch back to the insecure mode
     ///   let mut ftp_client = ftp_client.into_insecure().await.unwrap();
@@ -206,7 +291,7 @@ impl FtpClient {
     /// use ftp_rs::FtpClient;
     ///
     /// async {
-    ///   let client = FtpClient::connect("172.25.82.139:21").await
+    ///   let client = FtpClient::connect("192.168.32.204:21").await
     ///                          .expect("Couldn't connect to the server...");
     ///   let s: &TcpStream = client.get_ref();
     /// };
@@ -363,6 +448,54 @@ impl FtpClient {
             .await?)
     }
 
+    /// A convenience method to send the FTP PORT command to the server, receive the reply, and return the reply code.
+    pub async fn port(&mut self, host: IpAddr, port: u16) -> Result<u32> {
+        let mut args = String::with_capacity(24);
+        args.push_str(host.to_string().replace('.', ",").as_str());
+        args.push_str(",");
+        args.push_str((port >> 8).to_string().as_str());
+        args.push_str(",");
+        args.push_str((port & 0xff).to_string().as_str());
+        Ok(self
+            .send_command(Command::PORT, Some(args.as_str()))
+            .await?)
+    }
+
+    /// A convenience method to send the FTP EPRT command to the server, receive the reply, and return the reply code.
+    /// * EPRT |1|132.235.1.2|6275|
+    /// * EPRT |2|1080::8:800:200C:417A|5282|
+    pub async fn eprt(&mut self, host: IpAddr, port: u16) -> Result<u32> {
+        let mut args = String::new();
+        let mut h = host.to_string();
+        let n = h.find("%").unwrap_or(0);
+        if n > 0 {
+            h = h.substring(0, n).to_string();
+        }
+        args.push_str("|");
+        match host {
+            IpAddr::V4(addr) => args.push_str("1"),
+            IpAddr::V6(addr) => args.push_str("2"),
+        }
+        args.push_str("|");
+        args.push_str(h.as_str());
+        args.push_str("|");
+        args.push_str(port.to_string().as_str());
+        args.push_str("|");
+        Ok(self
+            .send_command(Command::EPRT, Some(args.as_str()))
+            .await?)
+    }
+
+    /// A convenience method to send the FTP MFMT command to the server, receive the reply, and return the reply code.
+    pub async fn mfmt(&mut self, pathname: &str, timeval: &str) -> Result<u32> {
+        Ok(self
+            .send_command(
+                Command::MFMT,
+                Some(format!("{} {}", timeval, pathname).as_str()),
+            )
+            .await?)
+    }
+
     /// Runs the PASV command.
     async fn pasv(&mut self) -> Result<SocketAddr> {
         self.send_command(Command::PASV, None).await?;
@@ -455,7 +588,7 @@ impl FtpClient {
     /// use tokio::io::{AsyncReadExt, BufStream};
     /// use std::io::Cursor;
     /// async {
-    ///   let mut conn = FtpClient::connect("172.25.82.139:21").await.unwrap();
+    ///   let mut conn = FtpClient::connect("192.168.32.204:21").await.unwrap();
     ///   conn.login("Doe", "mumble").await.unwrap();
     ///   let mut reader = Cursor::new("hello, world!".as_bytes());
     ///   conn.put("retr.txt", &mut reader).await.unwrap();
@@ -493,7 +626,7 @@ impl FtpClient {
     /// use ftp_rs::{FtpClient, FtpError};
     /// use std::io::Cursor;
     /// async {
-    ///     let mut conn = FtpClient::connect("172.25.82.139:21").await?;
+    ///     let mut conn = FtpClient::connect("192.168.32.204:21").await?;
     ///     conn.login("Doe", "mumble").await?;
     ///     let mut reader = Cursor::new("hello, world!".as_bytes());
     ///     conn.put("simple_retr.txt", &mut reader).await?;
